from datetime import datetime

from sqlalchemy.schema import Column
from sqlalchemy.types import DateTime, Integer

from sqlalchemy.ext.declarative import declarative_base, declared_attr


Base = declarative_base()


class BaseMixin:

    id = Column(Integer, primary_key=True)

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

    def __json_data__(self):
        data = {}
        for c in self.__table__.columns:
            data[c.name] = getattr(self, c.name)
        return data


class TimestampMixin:

    created_at = Column(DateTime, nullable=False, default=datetime.now)
    updated_at = Column(DateTime, onupdate=datetime.now)



import os

from tangled.decorators import cached_property
from tangled.web.events import TemplateContextCreated

from . import model


def include(app):
    # The theme can be specified as...
    #     - an absolute file system path
    #     - an asset path (e.g.: 'tangled.site:themes/default')
    #     - relative to tangled.site:themes (e.g.: 'default')
    theme = app.settings.get('site.theme', 'default')

    if os.path.isabs(theme):
        theme_directory = theme
    else:
        if ':' in theme:
            theme_directory = theme
        else:
            theme_directory = 'tangled.site:themes/{}'.format(theme)

    static_directory = os.path.join(theme_directory, 'static')
    app.mount_static_directory('static', static_directory)

    template_directory = os.path.join(theme_directory, 'templates')
    app.settings['mako.lookup.directories'] = template_directory

    # Home
    app.mount_resource('home', '.resources.home:Home', '/')

    # Auth
    app.mount_resource('sign-up', '.resources.auth:SignUp', '/sign-up')
    app.mount_resource('sign-in', '.resources.auth:SignIn', '/sign-in')
    app.mount_resource('sign-out', '.resources.auth:SignOut', '/sign-out')

    # Admin
    admin = app.mount_resource('admin', '.resources.admin:Admin', '/admin')
    admin.mount('users', 'users', method_name='users')
    admin.mount('meta', 'meta', factory='.resources.admin:Meta')

    # Users
    app.mount_resource('user', '.resources.user:User', '/user/<id>')
    app.mount_resource('profile', '.resources.user:Profile', '/profile')

    # Entries
    entries_settings = app.get_settings(prefix='site.entries.')
    path = entries_settings['path']
    name = path.strip('/')
    app.mount_resource(name, '.resources.entry:Entries', path)

    app.mount_resource('entries', '.resources.entry:Entries', '/entries')
    app.mount_resource(
        'new_entry', '.resources.entry:NewEntry', '/entries/new')

    # Entry
    app.mount_resource('entry', '.resources.entry:Entry', '/entry/<id>')
    app.mount_resource(
        'edit_entry', '.resources.entry:EditEntry', '/entry/<id>/edit')

    app.mount_resource(
        'page', '.resources.entry:Entry', '/<id>', methods='GET')

    def update_template_context(event):
        request = event.request
        q = request.db_session.query(model.Entry)
        q = q.filter_by(published=True, is_page=True)
        event.context['pages'] = q.all()
        event.context['user'] = request.user

    app.add_subscriber(TemplateContextCreated, update_template_context)

    @cached_property
    def user(request):
        user_id = request.authenticator.user_id
        if user_id is None:
            return None
        q = request.db_session.query(model.User)
        return q.get(user_id)

    app.add_request_attribute(user)

    def format_datetime(datetime):
        if not datetime:
            return datetime
        day_of_month = datetime.strftime('%b').lstrip('0')
        hour = datetime.strftime('%I').lstrip('0')
        am_pm = datetime.strftime('%p').lower()
        return datetime.strftime(
            '%a, {d} %d, %Y at {I}:%M{p}'
            .format(d=day_of_month, I=hour, p=am_pm))

    app.add_helper(format_datetime, static=True)

    app.load_config('.resources')



import logging

import bcrypt

from .model import User

from tangled.util import constant_time_compare


log = logging.getLogger(__name__)


def user_id_validator(request, user_id):
    """Ensure user (still) exists and return canonical ID.

    This is called by the registered Authenticator. ``user_id`` *must*
    be an integer, because that's what we store in the session as the
    user's canonical ID.

    """
    if not isinstance(user_id, int):
        log.error('User ID was not an int: {}'.format(int))
        request.abort(400)
    return request.db_session.query(User.id).filter_by(id=user_id).scalar()


def has_permission(request, user_id, permission):
    """Check to see if the user has the specified ``permission``.

    This is called by the registered Authorizer. ``user_id`` *must* be
    an integer, because that's what we store in the session as the
    user's canonical ID.

    """
    user = request.db_session.query(User).get(user_id)
    return user is not None and user.has_permission(permission)


def get_canonical_user_id(request, user_id):
    """Get canonical user ID.

    ``user_id`` can be a username or an email address. ``None`` is
    returned to indicate that the user doesn't exit.

    """
    if user_id is None:
        return None
    user_id = user_id.strip().lower()
    q = request.db_session.query(User.id)
    # Check by email
    canonical_id = q.filter_by(email=user_id).scalar()
    if canonical_id is not None:
        log.debug('Found user {} by email address'.format(canonical_id))
        return canonical_id
    # Check by username
    canonical_id = q.filter_by(username=user_id).scalar()
    if canonical_id is not None:
        log.debug('Found user {} by username'.format(canonical_id))
        return canonical_id
    log.debug('User does not exist: {}'.format(user_id))


def check_credentials(request, user_id, plain_text_password):
    """Check credentials.

    ``user_id`` may be a username or an email address.

    Returns the canonical user ID if the user exists and the credentials
    are valid. Returns ``None`` otherwise.

    """
    user_id = get_canonical_user_id(request, user_id)
    if user_id is None:
        log.debug('Unknown user: {}'.format(user_id))
        return None
    q = request.db_session.query(User.password)
    q = q.filter_by(id=user_id)
    hashed_password = q.scalar()
    if passwords_equal(plain_text_password, hashed_password):
        log.debug('Valid credentials')
        return user_id
    log.debug('Invalid credentials')


def hash_password(plain_text_password, salt=None):
    """Convert a plain text password to salted/hashed form."""
    plain_text_password = plain_text_password.encode('utf-8')
    if salt is None:
        salt = bcrypt.gensalt()
    hashed_password = bcrypt.hashpw(plain_text_password, salt)
    return hashed_password


def passwords_equal(plain_text_password, hashed_password):
    """Compare a plain text password to a hashed password."""
    plain_text_password_hashed = hash_password(
        plain_text_password, hashed_password)
    return constant_time_compare(plain_text_password_hashed, hashed_password)


def sign_up(request, email, plain_text_password, username=None):
    if get_canonical_user_id(request, email):
        log.debug('User exists with email: {}'.format(email))
        return None
    if get_canonical_user_id(request, username):
        log.debug('User exists with username: {}'.format(username))
        return None
    user = User(email=email, password=plain_text_password, username=username)
    request.db_session.add(user)
    request.db_session.flush([user])
    return user.id



from sqlalchemy import create_engine
from sqlalchemy.orm.session import sessionmaker

from tangled.abcs import ACommand

from .model import Base, Entry
from .model.user import User, Role, Permission


class Command(ACommand):

    @classmethod
    def configure(cls, parser):
        subparsers = parser.add_subparsers()
        init_db_parser = subparsers.add_parser('init_db')
        init_db_parser.add_argument('url')
        init_db_parser.add_argument('-u', '--username', default='admin')
        init_db_parser.add_argument(
            '-e', '--email', default='admin@example.com')
        init_db_parser.add_argument('-p', '--password', default='hunter2')
        init_db_parser.set_defaults(runner='init_db')

    def run(self):
        runner = getattr(self.args, 'runner', None)
        if runner:
            getattr(self, runner)()
        else:
            self.parser.print_help()

    def init_db(self):
        try:
            answer = input(
                'Are you sure you want to destroy existing data? '
                'Type "yes" to continue. ')
        except KeyboardInterrupt:
            self.exit('\nAborted.')

        if answer != 'yes':
            self.exit('Aborted.')

        username = self.args.username
        email = self.args.email
        password = self.args.password

        engine = create_engine(self.args.url)
        Base.metadata.drop_all(engine)
        Base.metadata.create_all(engine)
        session = sessionmaker(bind=engine)()

        sudo_permission = Permission(key='sudo')

        sudo_role = Role(
            key='sudo',
            description='Superuser; can do *anything*',
            permissions=[sudo_permission],
        )

        user = User(
            username=username,
            email=email,
            password=password,
            roles=[sudo_role],
        )

        home_entry = Entry(
            slug='home',
            title='Home',
            content='This is the home page.',
            published=True,
            is_page=True,
        )

        entry = Entry(
            slug='an-entry',
            title='This is an entry',
            content='It contains some content',
            published=True,
        )

        session.add_all([user, home_entry, entry])
        session.commit()
        session.close()

        print(
            'Created superuser {}/{} with password {}'
            .format(username, email, password))



from tangled.web import Resource, config


TITLES = {
    400: "Bad request",
    403: "You're not allowed to do that",
    404: "That page wasn't found",
    405: "That operation isn't supported",
    500: 'Something unexpected happened',
}


DETAILS = {

}


class Error(Resource):

    @config('text/html', template='error.mako')
    def GET(self):
        request = self.request
        response = request.response
        if response.status_code == 401:
            came_from = request.url
            location = request.make_url(
                '/sign-in', query={'came_from': came_from})
            request.abort(307, location=location)
        else:
            return {
                'title': TITLES.get(response.status_code, 'Error'),
                'detail': DETAILS.get(response.status_code, '')
            }



from sqlalchemy import event
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Table, Column, ForeignKey
from sqlalchemy.types import LargeBinary, String

from .base import Base, BaseMixin


class User(Base, BaseMixin):

    email = Column(String(length=100), nullable=False, unique=True)
    name = Column(String(length=100))
    username = Column(String(length=100), unique=True)
    password = Column(LargeBinary(60), nullable=False)

    roles = relationship('Role', secondary='user_role')

    @property
    def permissions(self):
        perms = set()
        for role in self.roles:
            perms.update(role.permissions)
        return perms

    def has_role(self, key):
        keys = set(role.key for role in self.roles)
        if 'sudo' in keys:
            return True
        return key in keys

    def has_permission(self, key):
        keys = set(permission.key for permission in self.permissions)
        if 'sudo' in keys:
            return True
        return key in keys


@event.listens_for(User.password, 'set', retval=True)
def hash_password(user, plain_text_password, old_hashed_password, event):
    import tangled.site.auth
    return tangled.site.auth.hash_password(plain_text_password)


class Permission(Base):

    __tablename__ = 'permission'

    key = Column(String(length=50), primary_key=True)
    description = Column(String(length=100))


class Role(Base):

    __tablename__ = 'role'

    key = Column(String(length=50), primary_key=True)
    description = Column(String(length=100))

    permissions = relationship(Permission, secondary='role_permission')


role_permission_table = Table(
    'role_permission', Base.metadata,
    Column('role_key', ForeignKey(Role.key), primary_key=True),
    Column('permission_key', ForeignKey(Permission.key), primary_key=True))


user_role_table = Table(
    'user_role', Base.metadata,
    Column('user_id', ForeignKey(User.id), primary_key=True),
    Column('role_key', ForeignKey(Role.key), primary_key=True))



from tangled.web import Resource

from ..resources.entry import Entries, Entry


class Home(Resource):

    def GET(self):
        app = self.app
        req = self.request
        home_slug = app.settings['site.home']
        if home_slug:
            resource = Entry(app, req, urlvars={'id': home_slug})
        else:
            resource = Entries(app, req)
        req.reset_resource_config()
        req.resource = resource
        return resource.GET()



from tangled.web import Resource, config
from tangled.web.abcs import AMountedResource

from ..model import User


@config('*/*', permission='admin')
class Admin(Resource):

    @config('text/html', template='admin/index.mako')
    def GET(self):
        mounted_resources = self.app.get_all(AMountedResource)
        resources = []
        for r in mounted_resources:
            if r.name.startswith('admin/'):
                resources.append(r)
        return {
            'resources': resources,
        }

    @config('text/html', template='admin/users.mako')
    def users(self):
        q = self.request.db_session.query(User)
        users = q.all()
        return {
            'users': users,
        }


@config('*/*', permission='sudo')
class Meta(Resource):

    """Metadata about the site that only superuses should see."""

    @config('text/html', template='admin/meta.mako')
    def GET(self):
        return {}



import logging

from tangled.web import Resource, config
from tangled.site import auth


log = logging.getLogger(__name__)


class SignUp(Resource):

    @config('text/html', template='auth/sign-up.mako')
    def GET(self):
        """Show sign-up form."""
        req = self.request
        if req.user:
            location = req.make_url('/profile')
            req.update_response(status=307, location=location)
        else:
            return {}

    @config('*/*', status=303, location='/profile', not_logged_in=True)
    def POST(self):
        req = self.request
        username = req.POST['username']
        email = req.POST['email']
        password = req.POST['password']
        user_id = auth.sign_up(req, email, password, username)
        if user_id is not None:
            req.authenticator.remember(user_id)
            req.expire_csrf_token()
        else:
            location = req.make_url('/sign-up')
            req.response.location = location
            req.flash('Could not sign up')


class SignIn(Resource):

    @config('text/html', template='/auth/sign-in.mako')
    def GET(self):
        if self.request.user:
            req = self.request
            location = req.make_url('/profile')
            req.update_response(status=307, location=location)
        else:
            return {}

    @config('*/*', status=303, location='CAME_FROM', not_logged_in=True)
    def POST(self):
        req = self.request
        user_id = req.POST['username']
        password = req.POST['password']
        canonical_user_id = auth.check_credentials(req, user_id, password)
        if canonical_user_id is not None:
            req.authenticator.remember(canonical_user_id)
            came_from = req.POST['came_from']
            if came_from.endswith('/sign-in'):
                del req.POST['came_from']
            req.expire_csrf_token()
        else:
            location = req.make_url('/sign-in')
            req.response.location = location
            req.flash('Could not sign in')


class SignOut(Resource):

    # TODO: Redirect to /logged-out
    @config('*/*', status=303, location='/', requires_authentication=True)
    def POST(self):
        req = self.request
        req.authenticator.forget()
        req.expire_csrf_token()
        req.session.clear()
        req.session.save()
        log.debug('Cleared session: {}'.format(req.session))
        req.flash('You have been signed out')



import re
from datetime import datetime

from markdown import Markdown

from sqlalchemy import event
from sqlalchemy.schema import Column
from sqlalchemy.types import Boolean, DateTime, String, Text

from .base import Base, BaseMixin, TimestampMixin


slug_re = re.compile(r'^[\w-]+$')


class Entry(Base, BaseMixin, TimestampMixin):

    slug = Column(String(length=100), nullable=False, unique=True)
    title = Column(String(length=100), nullable=False)
    content = Column(Text, nullable=False)
    content_html = Column(Text, nullable=False)
    is_page = Column(Boolean, default=False)
    published = Column(Boolean, default=False)
    published_at = Column(DateTime)


@event.listens_for(Entry.slug, 'set')
def on_set_slug(entry, value, *_):
    if not slug_re.search(value):
        raise ValueError('Slug may contain only word characters and dash')


@event.listens_for(Entry.content, 'set')
def set_content_html(entry, value, *_):
    """Set content_html from content whenever content is set."""
    entry.content_html = Markdown().convert(value)


@event.listens_for(Entry.published, 'set')
def on_set_published(entry, published, already_published, _):
    if not published:
        entry.published_at = None
    elif not already_published:
        entry.published_at = datetime.now()



from sqlalchemy.orm.exc import NoResultFound

from tangled.web import Resource, config

from .. import model


class Entries(Resource):

    @config('text/html', template='entry/entries.mako')
    def GET(self):
        session = self.request.db_session
        entries = session.query(model.Entry).all()
        return {
            'entries': entries,
        }

    @config('*/*', permission='create_entry')
    @config('text/html', status=303)
    def POST(self):
        """Create a new entry."""
        req = self.request
        content = req.POST['content']
        kwargs = dict(
            slug=req.POST['slug'],
            title=req.POST['title'],
            published=('published' in req.POST),
            is_page=('is_page' in req.POST),
            content=content,
        )
        new_entry = model.Entry(**kwargs)
        session = req.db_session
        session.add(new_entry)
        session.flush()
        location = req.resource_url('entry', {'id': new_entry.id})
        req.response.location = location


@config('*/*', permission='create_entry')
class NewEntry(Resource):

    @config('text/html', template='entry/new.mako')
    def GET(self):
        return {
            'entry': model.Entry(),
        }


class Entry(Resource):

    @config('text/html', template='entry/entry.mako')
    def GET(self):
        entry_id = self.urlvars['id']
        session = self.request.db_session
        user = self.request.user
        q = session.query(model.Entry)
        if not (user and user.has_permission('edit_entry')):
            q = q.filter_by(published=True)
        if entry_id.isdecimal():
            try:
                entry = q.filter_by(id=entry_id).one()
            except NoResultFound:
                pass
            else:
                return {
                    'entry': entry,
                }
        try:
            entry = q.filter_by(slug=entry_id).one()
        except NoResultFound:
            self.request.abort(404)
        return {
            'entry': entry,
        }

    @config('*/*', permission='edit_entry')
    @config('text/html', status=303)
    def PUT(self):
        req = self.request
        entry = self.GET()['entry']
        entry.slug = req.POST['slug']
        entry.title = req.POST['title']
        entry.published = 'published' in req.POST
        entry.is_page = 'is_page' in req.POST
        entry.content = req.POST['content']
        req.response.location = self.url()

    @config('*/*', permission='delete_entry')
    @config('text/html', status=303)
    def DELETE(self):
        entry = self.GET()['entry']
        self.request.db_session.delete(entry)
        location = self.request.resource_url('entries')
        self.request.response.location = location


@config('*/*', permission='edit_entry')
class EditEntry(Resource):

    @config('text/html', template='entry/edit.mako')
    def GET(self):
        resource = Entry(self.app, self.request, urlvars=self.urlvars)
        entry = resource.GET()['entry']
        return {
            'entry': entry,
        }



from setuptools import setup


setup(
    name='tangled.site',
    version='0.1a5',
    description='Simple site/blog/cms',
    long_description=open('README.rst').read(),
    url='http://tangledframework.org/',
    download_url='https://github.com/TangledWeb/tangled.site/tags',
    author='Wyatt Baldwin',
    author_email='self@wyattbaldwin.com',
    packages=[
        'tangled',
        'tangled.site',
        'tangled.site.model',
        'tangled.site.resources',
        'tangled.site.tests',
    ],
    include_package_data=True,
    install_requires=[
        'tangled.mako>=0.1a3',
        'tangled.sqlalchemy>=0.1a5',
        'tangled.web>=0.1a10',
        'bcrypt>=3.1.3',
        'Markdown>=2.6.8',
        'SQLAlchemy>=1.1.6',
    ],
    extras_require={
        'dev': [
            'tangled.web[dev]>=0.1a10',
        ],
    },
    entry_points="""
    [tangled.scripts]
    site = tangled.site.command

    """,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
)



from tangled.web import Resource, config

from .. import auth, model


@config('*/*', requires_authentication=True)
class User(Resource):

    @config('*/*', status=303, location='REFERER')
    def PUT(self):
        req = self.request
        user = req.user

        authorized = (
            user.id == int(self.urlvars['id']) or
            user.has_permission('update_user'))
        if not authorized:
            req.abort(403)

        params = self.app.get_settings(req.POST, 'user.')
        if not params:
            req.abort(400)

        if 'password' in params:
            current_password = params.pop('current_password')
            if not auth.passwords_equal(current_password, user.password):
                req.flash('Incorrect current password', 'error')
                return

            new_password = params.pop('password')
            confirm_password = params.pop('confirm_password')

            if new_password != confirm_password:
                req.flash("Passwords don't match", 'error')
                return

            if len(new_password) < 4:
                req.flash('That password is too short', 'error')
                return

            if not auth.passwords_equal(new_password, user.password):
                user.password = new_password
                req.flash('Password changed')

        if 'username' in params:
            username = params.pop('username')
            q = req.db_session.query(model.User.id)
            q = q.filter_by(username=username)
            user_id_for_username = q.scalar()
            if user_id_for_username is None:
                req.flash('Username updated')
                user.username = username
            elif user_id_for_username != user.id:
                req.flash('Username unavailable', 'error')

        if 'email' in params:
            email = params.pop('email')
            q = req.db_session.query(model.User.id)
            q = q.filter_by(email=email)
            user_id_for_email = q.scalar()
            if user_id_for_email is None:
                req.flash('Email address updated')
                user.email = email
            elif user_id_for_email != user.id:
                req.flash('Email address unavailable', 'error')

        # Set all other attributes directly
        for k, v in params.items():
            setattr(user, k, v)
            req.flash('{} updated'.format(k.title()))

    @config('*/*', permission='delete_user', status=303, location='REFERER')
    def DELETE(self):
        req = self.request
        user = req.db_session.query(model.User).get(self.urlvars['id'])
        if user.id == req.user.id:
            req.flash(
                "You can't delete your own account while you are signed in")
            req.abort(400)
        else:
            req.db_session.delete(user)


@config('*/*', requires_authentication=True)
class Profile(Resource):

    @config('text/html', template='profile.mako')
    def GET(self):
        return {}
